/*
 * Copyright (c) 2008, Ben L. Titzer
 * See license.txt for details.
 *
 * Created: May 19, 2008
 */
package aeneas.jvm;

import java.io.*;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.InvocationTargetException;

/**
 * The <code>V3System</code> class implements system calls for virgil programs compiled
 * to Java bytecode.
 *
 * @author Ben L. Titzer
 */
public class V3System {
    private static final int PRINT_SIZE = 128;
    private static final int MAX_FILES = 128;

    public static final File[] files;
    public static final FileInputStream[] fileInput;
    public static final FileOutputStream[] fileOutput;

    static {
        files = new File[MAX_FILES];
        fileInput = new FileInputStream[MAX_FILES];
        fileOutput = new FileOutputStream[MAX_FILES];

    }

    public static void main(String[] args) throws Exception {
        if (args.length < 1) {
            System.out.println("No Virgil program specified.");
            return;
        }
        Method mainMethod = findVirgilEntrypoint(getJavaClassName(args[0]));
        Class[] argumentTypes = mainMethod.getParameterTypes();
        mainMethod.setAccessible(true);
        if (argumentTypes.length == 1 && argumentTypes[0] == byte[][].class) {
            // invoke the main method with an array of byte arrays
            invokeMainMethod(mainMethod, new Object[] {convertArgsToByteArrays(args)});
        } else if (argumentTypes.length == 1 && argumentTypes[0] == String[].class) {
            // invoke the main method with an array of byte arrays
            invokeMainMethod(mainMethod, new Object[] {trimArgs(args)});
        } else if (argumentTypes.length == 1 && argumentTypes[0] == int.class) {
            // invoke the main method with a single integer
            if (args.length < 2) {
                System.out.println("Usage: " + args[0] + " <int>");
            } else {
                invokeMainMethod(mainMethod, new Object[] {Integer.parseInt(args[1])});
            }
        } else {
            System.out.println("Invalid main entrypoint signature: " + mainMethod.getParameterTypes());
        }
    }

    private static String[] trimArgs(String[] args) {
        String[] newArgs = new String[args.length - 1];
        for (int i = 0; i < newArgs.length; i++) {
            newArgs[i] = args[i + 1];
        }
        return newArgs;
    }

    private static byte[][] convertArgsToByteArrays(String[] args) {
        byte[][] newArgs = new byte[args.length - 1][];
        for (int i = 0; i < newArgs.length; i++) {
            newArgs[i] = args[i + 1].getBytes();
        }
        return newArgs;
    }

    private static void invokeMainMethod(Method mainMethod, Object[] arguments) throws Exception {
        try {
            mainMethod.invoke(null, arguments);
        } catch (InvocationTargetException e) {
            final Throwable cause = e.getTargetException();
            final StackTraceElement[] stackTraceElements = cause.getStackTrace();
            final String message = cause.getMessage();
            final String demangled = demangleString(message);
            System.out.print(cause.getClass().getName());
            System.out.print(": ");
            System.out.print(message);
            if (!message.equals(demangled)) {
                System.out.print(" (" + demangled + ")");
            }
            System.out.println();
            for (StackTraceElement stackTraceElement : stackTraceElements) {
                final String trace = stackTraceElement.toString();
                final String demangledTrace = demangleString(trace);
                System.out.print("\tin ");
                System.out.println(trace);
                if (!trace.equals(demangledTrace)) {
                    System.out.print("\t\t(");
                    System.out.print(demangledTrace);
                    System.out.println(")");
                }
            }
            throw e;
        }
    }

    private static String demangleString(String string) {
        if (string == null) return null;
        StringBuffer buffer = new StringBuffer(string.length() + 20);
        final int max = string.length();
        for (int i = 0; i < max; i++) {
            char ch = string.charAt(i);
            if (ch == '$' && i < max - 1) {
                i++;
                char n = string.charAt(i);
                switch (n) {
                    case 'A':
                        buffer.append(" -> ");
                        break;
                    case 'C':
                        buffer.append(',');
                        break;
                    case '_':
                        buffer.append(' ');
                        break;
                    case 'P':
                        buffer.append('(');
                        break;
                    case 'Q':
                        buffer.append(')');
                        break;
                    case 'L':
                        buffer.append('<');
                        break;
                    case 'R':
                        buffer.append('>');
                        break;
                    case 'D':
                        buffer.append('.');
                        break;
                    default:
                        buffer.append(ch);
                        i--;
                }
            } else {
                buffer.append(ch);
            }
        }
        return buffer.toString();
    }

    public static Method findVirgilEntrypoint(String className) throws Exception {
        Class testClass;
        try {
            testClass = Class.forName(className);
        } catch (ClassNotFoundException e) {
            throw new Exception("could not find class " + className);
        }
        for (Method method : testClass.getDeclaredMethods()) {
            if ((method.getModifiers() & Modifier.STATIC) != 0 && "main".equals(method.getName()))
                return method;
        }
        throw new Exception("could not find main method in class " + className);
    }


    public static String getJavaClassName(String className) {
        int ext = className.indexOf(".v3");
        if (ext > 0 && ext == className.length() - 3) className = className.substring(0, ext);
        int sep = className.lastIndexOf(File.separatorChar);
        if (sep > 0) className = className.substring(sep + 1);
        return "V3K_" + className;
    }

    public static void putc(byte ch) {
        System.out.print((char) ch);
    }

    public static void puts(byte[] str) {
        for (int i = 0; i < str.length; i += PRINT_SIZE) {
            int len = PRINT_SIZE;
            if (i + len >= str.length) len = str.length - i;
            System.out.write(str, i, len);
        }
    }

    public static void puti(int i) {
        System.out.print(i);
    }

    public static byte fileRead(int fd) throws IOException {
        if (files[fd] != null && fileInput[fd] != null) {
            return (byte)fileInput[fd].read();
        }
        return 0;
    }

    public static void fileWrite(int fd, byte ch) throws IOException {
        if (files[fd] != null && fileOutput[fd] != null) {
            fileOutput[fd].write(ch);
        }
    }

    public static int fileOpen(byte[] name, boolean input) {
        for (int i = 0; i < files.length; i++) {
            if (files[i] == null) {
                File file = new File(new String(name));
                files[i] = file;
                try {
                    if (input) {
                        fileInput[i] = new FileInputStream(file);
                    } else {
                        fileOutput[i] = new FileOutputStream(file);
                    }
                } catch (FileNotFoundException e) {
                    files[i] = null;
                    return -1;
                }
                return i;
            }
        }
        return -1;
    }

    public static int fileLeft(int fd) throws IOException {
        if (files[fd] != null && fileInput[fd] != null) {
            return fileInput[fd].available();
        }
        return 0;
    }

    public static void fileClose(int fd) {
        if (files[fd] != null) {
            try {
                if (fileInput[fd] != null) {
                    fileInput[fd].close();
                }
                if (fileOutput[fd] != null) {
                    fileOutput[fd].close();
                }
                files[fd] = null;
                fileInput[fd] = null;
                fileOutput[fd] = null;
            } catch (IOException e) {
                // do nothing.
            }
        }
    }

    public static byte[] fileLoad(byte[] fname) {
        try {
            File file = new File(new String(fname));
            FileInputStream fis = new FileInputStream(file);
            byte[] buffer = new byte[(int)file.length()];
            int pos = 0;
            while (pos < buffer.length) {
                pos += fis.read(buffer, pos, buffer.length - pos);
            }
            fis.close();
            return buffer;
        } catch (IOException e) {
            return null;
        }
    }

    public static void error(byte[] type, byte[] message) throws Exception {
        throw new Exception(new String(type) + ": " + new String(message));
    }

    public static boolean equals(Object a, Object b) {
        return a == b || (a == null ? b.equals(a) : a.equals(b));
    }
}
